/*
 * (c)2017, Cris Luengo.
 * Based on original DIPlib code: (c)1995-2014, Delft University of Technology.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "diplib.h"
#include "diplib/display.h"
#include "diplib/lookup_table.h"

namespace dip {

namespace {

constexpr dip::uint colorMapLength = 256;
constexpr dip::uint nRGB = 3;

// Generated in MATLAB using:
//    g = linspace(0,255,256)';
//    g = [g,g,g]
constexpr std::array< uint8, colorMapLength * nRGB > greyColorMap =
      {{
             0, 0, 0,
             1, 1, 1,
             2, 2, 2,
             3, 3, 3,
             4, 4, 4,
             5, 5, 5,
             6, 6, 6,
             7, 7, 7,
             8, 8, 8,
             9, 9, 9,
             10, 10, 10,
             11, 11, 11,
             12, 12, 12,
             13, 13, 13,
             14, 14, 14,
             15, 15, 15,
             16, 16, 16,
             17, 17, 17,
             18, 18, 18,
             19, 19, 19,
             20, 20, 20,
             21, 21, 21,
             22, 22, 22,
             23, 23, 23,
             24, 24, 24,
             25, 25, 25,
             26, 26, 26,
             27, 27, 27,
             28, 28, 28,
             29, 29, 29,
             30, 30, 30,
             31, 31, 31,
             32, 32, 32,
             33, 33, 33,
             34, 34, 34,
             35, 35, 35,
             36, 36, 36,
             37, 37, 37,
             38, 38, 38,
             39, 39, 39,
             40, 40, 40,
             41, 41, 41,
             42, 42, 42,
             43, 43, 43,
             44, 44, 44,
             45, 45, 45,
             46, 46, 46,
             47, 47, 47,
             48, 48, 48,
             49, 49, 49,
             50, 50, 50,
             51, 51, 51,
             52, 52, 52,
             53, 53, 53,
             54, 54, 54,
             55, 55, 55,
             56, 56, 56,
             57, 57, 57,
             58, 58, 58,
             59, 59, 59,
             60, 60, 60,
             61, 61, 61,
             62, 62, 62,
             63, 63, 63,
             64, 64, 64,
             65, 65, 65,
             66, 66, 66,
             67, 67, 67,
             68, 68, 68,
             69, 69, 69,
             70, 70, 70,
             71, 71, 71,
             72, 72, 72,
             73, 73, 73,
             74, 74, 74,
             75, 75, 75,
             76, 76, 76,
             77, 77, 77,
             78, 78, 78,
             79, 79, 79,
             80, 80, 80,
             81, 81, 81,
             82, 82, 82,
             83, 83, 83,
             84, 84, 84,
             85, 85, 85,
             86, 86, 86,
             87, 87, 87,
             88, 88, 88,
             89, 89, 89,
             90, 90, 90,
             91, 91, 91,
             92, 92, 92,
             93, 93, 93,
             94, 94, 94,
             95, 95, 95,
             96, 96, 96,
             97, 97, 97,
             98, 98, 98,
             99, 99, 99,
             100, 100, 100,
             101, 101, 101,
             102, 102, 102,
             103, 103, 103,
             104, 104, 104,
             105, 105, 105,
             106, 106, 106,
             107, 107, 107,
             108, 108, 108,
             109, 109, 109,
             110, 110, 110,
             111, 111, 111,
             112, 112, 112,
             113, 113, 113,
             114, 114, 114,
             115, 115, 115,
             116, 116, 116,
             117, 117, 117,
             118, 118, 118,
             119, 119, 119,
             120, 120, 120,
             121, 121, 121,
             122, 122, 122,
             123, 123, 123,
             124, 124, 124,
             125, 125, 125,
             126, 126, 126,
             127, 127, 127,
             128, 128, 128,
             129, 129, 129,
             130, 130, 130,
             131, 131, 131,
             132, 132, 132,
             133, 133, 133,
             134, 134, 134,
             135, 135, 135,
             136, 136, 136,
             137, 137, 137,
             138, 138, 138,
             139, 139, 139,
             140, 140, 140,
             141, 141, 141,
             142, 142, 142,
             143, 143, 143,
             144, 144, 144,
             145, 145, 145,
             146, 146, 146,
             147, 147, 147,
             148, 148, 148,
             149, 149, 149,
             150, 150, 150,
             151, 151, 151,
             152, 152, 152,
             153, 153, 153,
             154, 154, 154,
             155, 155, 155,
             156, 156, 156,
             157, 157, 157,
             158, 158, 158,
             159, 159, 159,
             160, 160, 160,
             161, 161, 161,
             162, 162, 162,
             163, 163, 163,
             164, 164, 164,
             165, 165, 165,
             166, 166, 166,
             167, 167, 167,
             168, 168, 168,
             169, 169, 169,
             170, 170, 170,
             171, 171, 171,
             172, 172, 172,
             173, 173, 173,
             174, 174, 174,
             175, 175, 175,
             176, 176, 176,
             177, 177, 177,
             178, 178, 178,
             179, 179, 179,
             180, 180, 180,
             181, 181, 181,
             182, 182, 182,
             183, 183, 183,
             184, 184, 184,
             185, 185, 185,
             186, 186, 186,
             187, 187, 187,
             188, 188, 188,
             189, 189, 189,
             190, 190, 190,
             191, 191, 191,
             192, 192, 192,
             193, 193, 193,
             194, 194, 194,
             195, 195, 195,
             196, 196, 196,
             197, 197, 197,
             198, 198, 198,
             199, 199, 199,
             200, 200, 200,
             201, 201, 201,
             202, 202, 202,
             203, 203, 203,
             204, 204, 204,
             205, 205, 205,
             206, 206, 206,
             207, 207, 207,
             208, 208, 208,
             209, 209, 209,
             210, 210, 210,
             211, 211, 211,
             212, 212, 212,
             213, 213, 213,
             214, 214, 214,
             215, 215, 215,
             216, 216, 216,
             217, 217, 217,
             218, 218, 218,
             219, 219, 219,
             220, 220, 220,
             221, 221, 221,
             222, 222, 222,
             223, 223, 223,
             224, 224, 224,
             225, 225, 225,
             226, 226, 226,
             227, 227, 227,
             228, 228, 228,
             229, 229, 229,
             230, 230, 230,
             231, 231, 231,
             232, 232, 232,
             233, 233, 233,
             234, 234, 234,
             235, 235, 235,
             236, 236, 236,
             237, 237, 237,
             238, 238, 238,
             239, 239, 239,
             240, 240, 240,
             241, 241, 241,
             242, 242, 242,
             243, 243, 243,
             244, 244, 244,
             245, 245, 245,
             246, 246, 246,
             247, 247, 247,
             248, 248, 248,
             249, 249, 249,
             250, 250, 250,
             251, 251, 251,
             252, 252, 252,
             253, 253, 253,
             254, 254, 254,
             255, 255, 255,
       }};

// Generated in MATLAB using:
//    g = linspace(0,255,256)';
//    g = [g,g,g];
//    g(1,:) = [0,0,255];
//    g(256,:) = [255,0,0]
constexpr std::array< uint8, colorMapLength * nRGB > saturationColorMap =
      {{
             0, 0, 255,
             1, 1, 1,
             2, 2, 2,
             3, 3, 3,
             4, 4, 4,
             5, 5, 5,
             6, 6, 6,
             7, 7, 7,
             8, 8, 8,
             9, 9, 9,
             10, 10, 10,
             11, 11, 11,
             12, 12, 12,
             13, 13, 13,
             14, 14, 14,
             15, 15, 15,
             16, 16, 16,
             17, 17, 17,
             18, 18, 18,
             19, 19, 19,
             20, 20, 20,
             21, 21, 21,
             22, 22, 22,
             23, 23, 23,
             24, 24, 24,
             25, 25, 25,
             26, 26, 26,
             27, 27, 27,
             28, 28, 28,
             29, 29, 29,
             30, 30, 30,
             31, 31, 31,
             32, 32, 32,
             33, 33, 33,
             34, 34, 34,
             35, 35, 35,
             36, 36, 36,
             37, 37, 37,
             38, 38, 38,
             39, 39, 39,
             40, 40, 40,
             41, 41, 41,
             42, 42, 42,
             43, 43, 43,
             44, 44, 44,
             45, 45, 45,
             46, 46, 46,
             47, 47, 47,
             48, 48, 48,
             49, 49, 49,
             50, 50, 50,
             51, 51, 51,
             52, 52, 52,
             53, 53, 53,
             54, 54, 54,
             55, 55, 55,
             56, 56, 56,
             57, 57, 57,
             58, 58, 58,
             59, 59, 59,
             60, 60, 60,
             61, 61, 61,
             62, 62, 62,
             63, 63, 63,
             64, 64, 64,
             65, 65, 65,
             66, 66, 66,
             67, 67, 67,
             68, 68, 68,
             69, 69, 69,
             70, 70, 70,
             71, 71, 71,
             72, 72, 72,
             73, 73, 73,
             74, 74, 74,
             75, 75, 75,
             76, 76, 76,
             77, 77, 77,
             78, 78, 78,
             79, 79, 79,
             80, 80, 80,
             81, 81, 81,
             82, 82, 82,
             83, 83, 83,
             84, 84, 84,
             85, 85, 85,
             86, 86, 86,
             87, 87, 87,
             88, 88, 88,
             89, 89, 89,
             90, 90, 90,
             91, 91, 91,
             92, 92, 92,
             93, 93, 93,
             94, 94, 94,
             95, 95, 95,
             96, 96, 96,
             97, 97, 97,
             98, 98, 98,
             99, 99, 99,
             100, 100, 100,
             101, 101, 101,
             102, 102, 102,
             103, 103, 103,
             104, 104, 104,
             105, 105, 105,
             106, 106, 106,
             107, 107, 107,
             108, 108, 108,
             109, 109, 109,
             110, 110, 110,
             111, 111, 111,
             112, 112, 112,
             113, 113, 113,
             114, 114, 114,
             115, 115, 115,
             116, 116, 116,
             117, 117, 117,
             118, 118, 118,
             119, 119, 119,
             120, 120, 120,
             121, 121, 121,
             122, 122, 122,
             123, 123, 123,
             124, 124, 124,
             125, 125, 125,
             126, 126, 126,
             127, 127, 127,
             128, 128, 128,
             129, 129, 129,
             130, 130, 130,
             131, 131, 131,
             132, 132, 132,
             133, 133, 133,
             134, 134, 134,
             135, 135, 135,
             136, 136, 136,
             137, 137, 137,
             138, 138, 138,
             139, 139, 139,
             140, 140, 140,
             141, 141, 141,
             142, 142, 142,
             143, 143, 143,
             144, 144, 144,
             145, 145, 145,
             146, 146, 146,
             147, 147, 147,
             148, 148, 148,
             149, 149, 149,
             150, 150, 150,
             151, 151, 151,
             152, 152, 152,
             153, 153, 153,
             154, 154, 154,
             155, 155, 155,
             156, 156, 156,
             157, 157, 157,
             158, 158, 158,
             159, 159, 159,
             160, 160, 160,
             161, 161, 161,
             162, 162, 162,
             163, 163, 163,
             164, 164, 164,
             165, 165, 165,
             166, 166, 166,
             167, 167, 167,
             168, 168, 168,
             169, 169, 169,
             170, 170, 170,
             171, 171, 171,
             172, 172, 172,
             173, 173, 173,
             174, 174, 174,
             175, 175, 175,
             176, 176, 176,
             177, 177, 177,
             178, 178, 178,
             179, 179, 179,
             180, 180, 180,
             181, 181, 181,
             182, 182, 182,
             183, 183, 183,
             184, 184, 184,
             185, 185, 185,
             186, 186, 186,
             187, 187, 187,
             188, 188, 188,
             189, 189, 189,
             190, 190, 190,
             191, 191, 191,
             192, 192, 192,
             193, 193, 193,
             194, 194, 194,
             195, 195, 195,
             196, 196, 196,
             197, 197, 197,
             198, 198, 198,
             199, 199, 199,
             200, 200, 200,
             201, 201, 201,
             202, 202, 202,
             203, 203, 203,
             204, 204, 204,
             205, 205, 205,
             206, 206, 206,
             207, 207, 207,
             208, 208, 208,
             209, 209, 209,
             210, 210, 210,
             211, 211, 211,
             212, 212, 212,
             213, 213, 213,
             214, 214, 214,
             215, 215, 215,
             216, 216, 216,
             217, 217, 217,
             218, 218, 218,
             219, 219, 219,
             220, 220, 220,
             221, 221, 221,
             222, 222, 222,
             223, 223, 223,
             224, 224, 224,
             225, 225, 225,
             226, 226, 226,
             227, 227, 227,
             228, 228, 228,
             229, 229, 229,
             230, 230, 230,
             231, 231, 231,
             232, 232, 232,
             233, 233, 233,
             234, 234, 234,
             235, 235, 235,
             236, 236, 236,
             237, 237, 237,
             238, 238, 238,
             239, 239, 239,
             240, 240, 240,
             241, 241, 241,
             242, 242, 242,
             243, 243, 243,
             244, 244, 244,
             245, 245, 245,
             246, 246, 246,
             247, 247, 247,
             248, 248, 248,
             249, 249, 249,
             250, 250, 250,
             251, 251, 251,
             252, 252, 252,
             253, 253, 253,
             254, 254, 254,
             255, 0,   0,
       }};

// Black + 14 repetitions of 19 unique colors.
// Greens are closer together perceptually, so we reduce the number of green entries in this map.
// Assuming gamma of 1.2: ((1/3)^(1/1.2))=102 ; ((2/3)^(1/1.2))=182 ; ((1/2)^(1/1.2))=143
// Data same as in /dipimage/private/label_colormap.m
constexpr std::array< uint8, colorMapLength * nRGB > labelColorMap =
      {{
             0,   0,   0,

             255, 0,   0,
             0,   255, 0, // green
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             //0,   255, 182, // greenish A
             0,   255, 143, // a different greenish color in between A and C
             182, 0,   255,
             255, 0,   182,
             //182, 255, 0, // greenish B
             143, 255, 0, // a different greenish color in between B and D
             0,   182, 255,
             255, 102, 0,
             //0,   255, 102, // greenish C
             102, 0,   255,
             255, 0,   102,
             //102, 255, 0, // greenish D
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             182, 0,   255,
             255, 0,   182,
             143, 255, 0,
             0,   182, 255,
             255, 102, 0,
             102, 0,   255,
             255, 0,   102,
             0,   102, 255,
             182, 182, 0,
             0,   182, 182,
             182, 0,   182,

             255, 0,   0,
             0,   255, 0,
             0,   0,   255,
             255, 255, 0,
             0,   255, 255,
             255, 0,   255,
             255, 182, 0,
             0,   255, 143,
             //182, 0,   255,
             //255, 0,   182,
             //143, 255, 0,
             //0,   182, 255,
             //255, 102, 0,
             //102, 0,   255,
             //255, 0,   102,
             //0,   102, 255,
             //182, 182, 0,
             //0,   182, 182,
             //182, 0,   182,

       }};

/*
 * The following colormaps are from colorcet:
 * http://peterkovesi.com/projects/colourmaps/index.html
 *
 * Copyright (c) 2017 Peter Kovesi
 * Centre for Exploration Targeting
 * The University of Western Australia
 * peter.kovesi at uwa edu au
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * The Software is provided "as is", without warranty of any kind.
 */

// Linear blue - magenta - yellow highly saturated colour map (CET-L08)
constexpr std::array< uint8, colorMapLength * nRGB > linearColorMap =
      {{
             0, 15, 93,
             1, 15, 94,
             1, 16, 96,
             1, 16, 98,
             1, 17, 99,
             2, 17, 101,
             2, 18, 103,
             3, 18, 104,
             4, 18, 106,
             5, 19, 107,
             6, 19, 109,
             7, 20, 111,
             8, 20, 112,
             10, 21, 114,
             12, 21, 115,
             13, 21, 117,
             15, 22, 118,
             17, 22, 119,
             18, 22, 121,
             20, 23, 122,
             22, 23, 124,
             24, 23, 125,
             25, 24, 126,
             27, 24, 128,
             29, 24, 129,
             31, 25, 130,
             33, 25, 131,
             35, 25, 133,
             37, 25, 134,
             39, 26, 135,
             41, 26, 136,
             43, 26, 137,
             46, 26, 138,
             48, 26, 139,
             50, 27, 140,
             52, 27, 141,
             55, 27, 142,
             57, 27, 143,
             59, 27, 143,
             62, 27, 144,
             64, 27, 144,
             67, 26, 145,
             70, 26, 145,
             72, 26, 146,
             75, 26, 146,
             78, 25, 146,
             81, 25, 146,
             84, 25, 146,
             86, 24, 146,
             89, 24, 146,
             92, 23, 146,
             94, 23, 146,
             97, 23, 146,
             99, 22, 146,
             102, 22, 146,
             104, 21, 146,
             107, 21, 146,
             109, 20, 146,
             111, 20, 146,
             114, 19, 146,
             116, 19, 146,
             118, 18, 145,
             121, 18, 145,
             123, 17, 145,
             125, 17, 145,
             127, 16, 145,
             129, 16, 145,
             131, 15, 144,
             134, 15, 144,
             136, 14, 144,
             138, 14, 144,
             140, 13, 143,
             142, 13, 143,
             144, 13, 143,
             146, 12, 142,
             148, 12, 142,
             150, 11, 142,
             152, 11, 141,
             154, 11, 141,
             156, 10, 140,
             158, 10, 140,
             160, 10, 140,
             162, 10, 139,
             163, 10, 139,
             165, 9, 138,
             167, 9, 138,
             169, 9, 138,
             171, 9, 137,
             173, 9, 137,
             174, 10, 136,
             176, 10, 136,
             178, 10, 136,
             180, 10, 135,
             182, 10, 135,
             183, 11, 134,
             185, 11, 134,
             187, 12, 133,
             189, 12, 133,
             190, 13, 133,
             192, 13, 132,
             194, 14, 132,
             195, 14, 131,
             197, 15, 131,
             199, 16, 130,
             200, 16, 130,
             202, 17, 129,
             204, 18, 129,
             205, 19, 128,
             207, 20, 128,
             208, 21, 127,
             210, 22, 127,
             211, 23, 126,
             213, 25, 125,
             214, 26, 125,
             215, 28, 124,
             217, 29, 123,
             218, 31, 123,
             220, 32, 122,
             221, 34, 121,
             222, 35, 120,
             223, 37, 120,
             224, 39, 119,
             226, 41, 118,
             227, 42, 117,
             228, 44, 116,
             229, 46, 115,
             230, 48, 114,
             231, 49, 114,
             232, 51, 113,
             233, 53, 112,
             234, 55, 111,
             235, 57, 110,
             236, 59, 108,
             237, 61, 107,
             237, 62, 106,
             238, 64, 105,
             239, 66, 104,
             240, 68, 103,
             240, 70, 102,
             241, 72, 101,
             242, 74, 100,
             242, 76, 98,
             243, 77, 97,
             243, 79, 96,
             244, 81, 95,
             244, 83, 94,
             245, 85, 93,
             245, 87, 92,
             246, 88, 92,
             246, 90, 91,
             247, 92, 90,
             247, 94, 89,
             247, 96, 88,
             248, 98, 87,
             248, 100, 86,
             248, 101, 85,
             248, 103, 85,
             248, 105, 84,
             249, 107, 83,
             249, 109, 82,
             249, 111, 82,
             249, 112, 81,
             249, 114, 80,
             249, 116, 79,
             249, 118, 79,
             249, 120, 78,
             249, 122, 77,
             249, 123, 77,
             249, 125, 76,
             249, 127, 76,
             249, 129, 75,
             249, 130, 74,
             249, 132, 74,
             249, 134, 73,
             249, 135, 72,
             249, 137, 72,
             249, 138, 71,
             249, 140, 70,
             249, 142, 70,
             249, 143, 69,
             249, 145, 68,
             249, 146, 68,
             249, 148, 67,
             249, 149, 66,
             249, 151, 66,
             250, 152, 65,
             250, 154, 65,
             250, 155, 64,
             250, 157, 63,
             250, 158, 63,
             250, 159, 62,
             250, 161, 61,
             251, 162, 61,
             251, 164, 60,
             251, 165, 59,
             251, 166, 59,
             251, 168, 58,
             252, 169, 57,
             252, 171, 57,
             252, 172, 57,
             252, 173, 56,
             252, 175, 56,
             252, 176, 56,
             252, 177, 56,
             252, 179, 56,
             252, 180, 56,
             252, 182, 56,
             253, 183, 56,
             253, 184, 56,
             253, 186, 56,
             253, 187, 56,
             253, 189, 56,
             253, 190, 56,
             253, 191, 56,
             253, 193, 57,
             253, 194, 57,
             252, 195, 57,
             252, 197, 57,
             252, 198, 58,
             252, 200, 58,
             252, 201, 58,
             252, 202, 59,
             252, 204, 59,
             252, 205, 59,
             252, 206, 60,
             252, 208, 60,
             252, 209, 61,
             252, 211, 61,
             251, 212, 62,
             251, 213, 62,
             251, 215, 63,
             251, 216, 63,
             251, 217, 64,
             251, 219, 64,
             251, 220, 65,
             250, 222, 65,
             250, 223, 66,
             250, 224, 66,
             250, 226, 67,
             250, 227, 67,
             249, 228, 68,
             249, 230, 69,
             249, 231, 69,
             249, 232, 70,
             248, 234, 70,
             248, 235, 71,
             248, 237, 72,
             248, 238, 72,
             247, 239, 73,
             247, 241, 74,
             247, 242, 74,
             247, 243, 75,
             246, 245, 76,
             246, 246, 76,
             246, 247, 77,
             245, 249, 78
       }};

// Linear-diverging blue - grey - red (CET-D08)
constexpr std::array< uint8, colorMapLength * nRGB > divergingColorMap =
      {{
             0, 42, 215,
             0, 43, 214,
             0, 43, 213,
             0, 44, 212,
             0, 44, 212,
             0, 45, 211,
             0, 45, 210,
             0, 46, 209,
             7, 46, 208,
             13, 47, 207,
             19, 47, 206,
             23, 48, 205,
             27, 48, 204,
             30, 49, 203,
             33, 49, 202,
             35, 50, 201,
             38, 50, 201,
             40, 51, 200,
             42, 51, 199,
             44, 52, 198,
             46, 52, 197,
             47, 53, 196,
             49, 53, 195,
             51, 54, 194,
             52, 54, 193,
             54, 55, 192,
             55, 55, 191,
             56, 56, 191,
             58, 56, 190,
             59, 57, 189,
             60, 57, 188,
             61, 57, 187,
             62, 58, 186,
             64, 58, 185,
             65, 59, 184,
             66, 59, 183,
             67, 60, 182,
             68, 60, 181,
             69, 61, 181,
             69, 61, 180,
             70, 62, 179,
             71, 62, 178,
             72, 63, 177,
             73, 63, 176,
             74, 64, 175,
             74, 64, 174,
             75, 64, 173,
             76, 65, 172,
             77, 65, 172,
             77, 66, 171,
             78, 66, 170,
             79, 67, 169,
             79, 67, 168,
             80, 68, 167,
             81, 68, 166,
             81, 69, 165,
             82, 69, 164,
             82, 70, 163,
             83, 70, 163,
             84, 70, 162,
             84, 71, 161,
             85, 71, 160,
             85, 72, 159,
             86, 72, 158,
             86, 73, 157,
             87, 73, 156,
             87, 74, 155,
             88, 74, 155,
             88, 74, 154,
             88, 75, 153,
             89, 75, 152,
             89, 76, 151,
             90, 76, 150,
             90, 77, 149,
             91, 77, 148,
             91, 78, 147,
             91, 78, 146,
             92, 78, 146,
             92, 79, 145,
             92, 79, 144,
             93, 80, 143,
             93, 80, 142,
             93, 81, 141,
             94, 81, 140,
             94, 82, 139,
             94, 82, 138,
             94, 82, 138,
             95, 83, 137,
             95, 83, 136,
             95, 84, 135,
             96, 84, 134,
             96, 85, 133,
             96, 85, 132,
             96, 86, 131,
             96, 86, 130,
             97, 86, 130,
             97, 87, 129,
             97, 87, 128,
             97, 88, 127,
             98, 88, 126,
             98, 89, 125,
             98, 89, 124,
             98, 89, 123,
             98, 90, 122,
             98, 90, 122,
             99, 91, 121,
             99, 91, 120,
             99, 92, 119,
             99, 92, 118,
             99, 92, 117,
             99, 93, 116,
             99, 93, 115,
             99, 94, 114,
             99, 94, 113,
             100, 95, 113,
             100, 95, 112,
             100, 95, 111,
             100, 96, 110,
             100, 96, 109,
             100, 97, 108,
             100, 97, 107,
             100, 98, 106,
             100, 98, 105,
             100, 98, 104,
             100, 99, 103,
             100, 99, 103,
             100, 100, 102,
             100, 100, 101,
             101, 100, 100,
             103, 100, 99,
             105, 100, 99,
             106, 100, 98,
             108, 100, 97,
             110, 100, 97,
             111, 99, 96,
             113, 99, 96,
             114, 99, 95,
             116, 99, 94,
             117, 99, 94,
             119, 99, 93,
             120, 99, 92,
             122, 98, 92,
             123, 98, 91,
             125, 98, 91,
             126, 98, 90,
             128, 98, 89,
             129, 98, 89,
             131, 97, 88,
             132, 97, 87,
             134, 97, 87,
             135, 97, 86,
             136, 97, 86,
             138, 96, 85,
             139, 96, 84,
             140, 96, 84,
             142, 96, 83,
             143, 96, 82,
             144, 95, 82,
             146, 95, 81,
             147, 95, 81,
             148, 95, 80,
             150, 94, 79,
             151, 94, 79,
             152, 94, 78,
             154, 94, 77,
             155, 93, 77,
             156, 93, 76,
             158, 93, 75,
             159, 93, 75,
             160, 92, 74,
             161, 92, 73,
             163, 92, 73,
             164, 91, 72,
             165, 91, 72,
             166, 91, 71,
             168, 91, 70,
             169, 90, 70,
             170, 90, 69,
             171, 90, 68,
             173, 89, 68,
             174, 89, 67,
             175, 89, 66,
             176, 88, 66,
             177, 88, 65,
             179, 88, 64,
             180, 87, 64,
             181, 87, 63,
             182, 86, 62,
             183, 86, 62,
             185, 86, 61,
             186, 85, 60,
             187, 85, 60,
             188, 84, 59,
             189, 84, 58,
             191, 84, 58,
             192, 83, 57,
             193, 83, 56,
             194, 82, 56,
             195, 82, 55,
             197, 81, 54,
             198, 81, 53,
             199, 80, 53,
             200, 80, 52,
             201, 79, 51,
             202, 79, 51,
             204, 78, 50,
             205, 78, 49,
             206, 77, 48,
             207, 77, 48,
             208, 76, 47,
             209, 76, 46,
             210, 75, 46,
             212, 74, 45,
             213, 74, 44,
             214, 73, 43,
             215, 73, 43,
             216, 72, 42,
             217, 71, 41,
             218, 71, 40,
             220, 70, 39,
             221, 69, 39,
             222, 69, 38,
             223, 68, 37,
             224, 67, 36,
             225, 66, 35,
             226, 65, 35,
             228, 65, 34,
             229, 64, 33,
             230, 63, 32,
             231, 62, 31,
             232, 61, 30,
             233, 60, 29,
             234, 59, 28,
             235, 58, 27,
             237, 58, 27,
             238, 56, 26,
             239, 55, 25,
             240, 54, 24,
             241, 53, 22,
             242, 52, 21,
             243, 51, 20,
             244, 50, 19,
             246, 48, 18,
             247, 47, 17,
             248, 46, 15,
             249, 44, 14,
             250, 43, 13,
             251, 41, 11,
             252, 40, 9,
             253, 38, 8,
             255, 36, 6,
             255, 34, 5,
             255, 32, 3,
             255, 30, 1,
             255, 27, 0,
             255, 25, 0
       }};

// Cyclic magenta - yellow - green - blue (CET-C2)
constexpr std::array< uint8, colorMapLength * nRGB > cyclicColorMap =
      {{
             239, 85, 242,
             241, 87, 240,
             242, 90, 239,
             243, 93, 237,
             245, 95, 235,
             246, 98, 233,
             246, 101, 231,
             247, 104, 229,
             248, 107, 226,
             249, 111, 224,
             249, 114, 222,
             250, 117, 219,
             250, 120, 217,
             250, 123, 214,
             251, 126, 211,
             251, 129, 209,
             251, 132, 206,
             251, 135, 203,
             251, 138, 201,
             252, 141, 198,
             252, 144, 195,
             252, 147, 192,
             252, 149, 190,
             252, 152, 187,
             252, 155, 184,
             252, 158, 181,
             252, 160, 179,
             252, 163, 176,
             252, 166, 173,
             252, 168, 170,
             252, 171, 167,
             252, 173, 165,
             252, 176, 162,
             251, 178, 159,
             251, 181, 156,
             251, 183, 153,
             251, 185, 150,
             252, 188, 147,
             252, 190, 144,
             252, 192, 141,
             252, 195, 138,
             252, 197, 135,
             252, 199, 132,
             252, 201, 129,
             252, 204, 126,
             252, 206, 123,
             252, 208, 120,
             252, 210, 117,
             252, 212, 113,
             252, 214, 110,
             252, 216, 107,
             252, 218, 103,
             252, 221, 100,
             251, 223, 96,
             251, 224, 92,
             251, 226, 88,
             250, 228, 85,
             250, 230, 81,
             249, 231, 77,
             248, 233, 73,
             247, 234, 69,
             245, 235, 65,
             244, 236, 61,
             242, 237, 57,
             241, 238, 53,
             239, 238, 50,
             237, 238, 46,
             234, 239, 43,
             232, 238, 40,
             230, 238, 37,
             227, 238, 35,
             225, 237, 33,
             222, 237, 31,
             219, 236, 29,
             216, 235, 27,
             214, 234, 26,
             211, 233, 25,
             208, 232, 24,
             205, 231, 24,
             202, 230, 23,
             199, 229, 22,
             196, 228, 22,
             193, 227, 21,
             190, 226, 21,
             187, 225, 21,
             184, 224, 20,
             181, 223, 20,
             178, 222, 19,
             175, 221, 19,
             172, 220, 19,
             169, 218, 18,
             166, 217, 18,
             163, 216, 18,
             160, 215, 17,
             157, 214, 17,
             154, 213, 17,
             151, 212, 16,
             148, 211, 16,
             145, 210, 16,
             141, 208, 15,
             138, 207, 15,
             135, 206, 15,
             132, 205, 14,
             129, 204, 14,
             125, 203, 14,
             122, 202, 13,
             119, 200, 13,
             115, 199, 13,
             112, 198, 13,
             108, 197, 12,
             105, 196, 12,
             101, 195, 12,
             98, 194, 12,
             94, 192, 12,
             90, 191, 12,
             86, 190, 12,
             83, 189, 13,
             79, 188, 14,
             75, 186, 15,
             72, 185, 16,
             68, 184, 18,
             65, 183, 20,
             62, 181, 22,
             59, 180, 25,
             56, 179, 27,
             54, 177, 30,
             52, 176, 33,
             51, 174, 37,
             50, 173, 40,
             50, 171, 44,
             50, 169, 47,
             50, 168, 51,
             51, 166, 55,
             53, 164, 59,
             54, 163, 63,
             56, 161, 66,
             57, 159, 70,
             59, 157, 74,
             60, 155, 78,
             62, 153, 82,
             63, 152, 86,
             64, 150, 89,
             65, 148, 93,
             66, 146, 97,
             67, 144, 100,
             68, 142, 104,
             68, 140, 107,
             69, 138, 111,
             69, 137, 114,
             69, 135, 118,
             69, 133, 121,
             69, 131, 125,
             69, 129, 128,
             68, 127, 132,
             68, 125, 135,
             67, 123, 138,
             66, 121, 142,
             65, 120, 145,
             64, 118, 148,
             63, 116, 151,
             62, 114, 155,
             60, 112, 158,
             59, 110, 161,
             57, 108, 164,
             56, 106, 167,
             54, 104, 170,
             52, 102, 173,
             51, 100, 176,
             49, 97, 179,
             48, 95, 182,
             47, 93, 185,
             45, 91, 187,
             44, 88, 190,
             43, 86, 193,
             42, 83, 195,
             42, 81, 198,
             41, 78, 200,
             40, 76, 203,
             39, 73, 205,
             39, 70, 208,
             38, 67, 210,
             38, 64, 213,
             37, 62, 215,
             37, 59, 217,
             36, 56, 219,
             36, 53, 222,
             37, 50, 224,
             37, 47, 226,
             38, 44, 228,
             40, 41, 229,
             41, 39, 231,
             44, 36, 233,
             46, 34, 234,
             49, 32, 236,
             52, 30, 237,
             55, 29, 238,
             59, 28, 239,
             62, 27, 240,
             66, 27, 241,
             69, 27, 242,
             73, 27, 242,
             77, 28, 243,
             80, 28, 243,
             84, 29, 244,
             87, 31, 244,
             90, 32, 245,
             94, 33, 245,
             97, 35, 246,
             100, 36, 246,
             103, 38, 246,
             106, 39, 247,
             109, 41, 247,
             112, 42, 247,
             115, 44, 248,
             117, 45, 248,
             120, 46, 248,
             123, 48, 249,
             126, 49, 249,
             128, 50, 249,
             131, 52, 250,
             134, 53, 250,
             137, 54, 250,
             140, 55, 250,
             143, 56, 251,
             145, 57, 251,
             148, 58, 251,
             152, 59, 251,
             155, 60, 251,
             158, 61, 251,
             161, 62, 251,
             164, 62, 251,
             167, 63, 251,
             171, 64, 251,
             174, 64, 251,
             177, 65, 251,
             181, 65, 251,
             184, 66, 251,
             187, 66, 251,
             191, 67, 251,
             194, 67, 251,
             197, 68, 251,
             200, 68, 250,
             203, 69, 250,
             207, 69, 250,
             210, 70, 250,
             213, 70, 250,
             216, 71, 249,
             219, 72, 249,
             221, 72, 249,
             224, 73, 248,
             227, 75, 248,
             229, 76, 247,
             231, 77, 246,
             233, 79, 245,
             236, 81, 244,
             237, 83, 243
       }};

}

/*
 * End of colormaps from colorcet.
 */

void ApplyColorMap(
      Image const& in,
      Image& out,
      String const& colorMap
) {
   uint8 const* values = nullptr;
   if(( colorMap == "grey" ) || ( colorMap == "gray" )) {
      values = greyColorMap.data();
   } else if( colorMap == "saturation" ) {
      values = saturationColorMap.data();
   } else if( colorMap == "linear" ) {
      values = linearColorMap.data();
   } else if( colorMap == "diverging" ) {
      values = divergingColorMap.data();
   } else if( colorMap == "cyclic" ) {
      values = cyclicColorMap.data();
   } else if( colorMap == "label" ) {
      values = labelColorMap.data();
   } else {
      DIP_THROW_INVALID_FLAG( colorMap );
   }
   Image im( values, { colorMapLength }, nRGB );
   LookupTable lut( im );
   DIP_STACK_TRACE_THIS( lut.Apply( in, out ));
   out.SetColorSpace( "sRGB" );
}

void Overlay(
      Image const& c_in,
      Image const& overlay,
      Image& out,
      Image::Pixel const& color
) {
   DIP_THROW_IF( !c_in.IsForged() || !overlay.IsForged(), E::IMAGE_NOT_FORGED );
   DIP_THROW_IF( overlay.Sizes() != c_in.Sizes(), E::SIZES_DONT_MATCH );
   DIP_THROW_IF( !overlay.IsScalar(), E::MASK_NOT_SCALAR );
   if( out.IsForged() && out.IsSingletonExpanded() ) {
      // This could happen if &out == &c_in.
      out.Strip();
   }
   Image in = c_in;
   if( !(overlay.DataType().IsBinary() && in.IsScalar() && color.IsScalar() )) {
      // This code is not run in the case that `overlay` is binary, and `in` and `color` are scalar. In this case,
      //    we create a grey-value output.
      if( in.IsScalar() ) {
         in.ExpandSingletonTensor( 3 );
         in.SetColorSpace( "sRGB" );
         if( out.Aliases( in )) {
            out.Strip(); // if &out == &in, we need a new bit of memory for out
         }
      } else {
         DIP_THROW_IF( in.TensorElements() != 3, "Input image must have 1 or 3 tensor elements" );
         DIP_THROW_IF( in.IsColor() && ( in.ColorSpace() != "sRGB" ), "Convert input image to sRGB color space first" );
         in.SetColorSpace( "sRGB" );
      }
   }
   if( in.DataType().IsBinary() ) {
      Convert( in, out, DT_UINT8 );
      out *= 255;
   } else {
      Copy( in, out );
   }
   if( overlay.DataType().IsBinary() ) {
      // A binary overlay
      DIP_THROW_IF( !color.IsScalar() && ( color.TensorElements() != 3 ), "Color must have 1 or 3 tensor elements" );
      DIP_STACK_TRACE_THIS( out.At( overlay ) = color );
      // out.At( overlay ).Fill( color );
   } else if( overlay.DataType().IsUnsigned() ) {
      // A labeled overlay
      Image mask = overlay > 0;
      Image labels = overlay.At( mask );
      DIP_STACK_TRACE_THIS( ApplyColorMap( labels, labels, "label" ));
      DIP_STACK_TRACE_THIS( out.At( mask ) = labels );
      // out.At( mask ).Copy( labels );
   } else {
      DIP_THROW( E::DATA_TYPE_NOT_SUPPORTED );
   }
}

} // namespace dip
